/**
 * @file SpritesCommon.h
 * \brief
 * Common header file for sprite functions
 */

#ifndef SpritesCommon_h
#define SpritesCommon_h

#define SPRITE_MASKED 1
#define SPRITE_UNMASKED 2
#define SPRITE_OVERWRITE 2
#define SPRITE_PLUS_MASK 3
#define SPRITE_IS_MASK 250
#define SPRITE_IS_MASK_ERASE 251
#define SPRITE_AUTO_MODE 255

#endif
